
/*
 *  Diverse Bristol audio routines.
 *  Copyright (c) by Nick Copeland <nickycopeland@hotmail.com> 1996,2006
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//#define DEBUG

#include <math.h>

#include "bristol.h"
#include "reverb.h"

#define DELAY		0
#define FEEDBACK	1
#define CROSSOVER	2
#define WETDRY		3
#define LEFT		4
#define RIGHT		5

#define OPNAME "Reverb"
#define OPDESCRIPTION "Stereo input/output reverb"
#define PCOUNT 6
#define IOCOUNT 4

#define VREVERB_LIN_IND 0
#define VREVERB_RIN_IND 1
#define VREVERB_LOUT_IND 2
#define VREVERB_ROUT_IND 3

/*
 * Reset any local memory information.
 */
static int destroy(bristolOP *operator)
{
#ifdef BRISTOL_DBG
	printf("destroy(%x)\n", operator);
#endif

	bristolfree(operator->specs);

	cleanup(operator);
}

/*
 * This is called by the frontend when a parameter is changed.
static int param(param, value)
 */
static int param(bristolOP *operator, bristolOPParams *param,
	unsigned char index, float value)
{
#ifdef DEBUG
	printf("checkParams(%f)\n", value);
#endif

	param->param[index].float_val = value;
}

/*
 * Reset any local memory information.
 */
static int reset(bristolOP *operator, bristolOPParams *param)
{
#ifdef BRISTOL_DBG
	printf("reset(%x)\n", operator);
#endif

	if (param->param[0].mem)
		bristolfree(param->param[0].mem);
	if (param->param[1].mem)
		bristolfree(param->param[1].mem);

	param->param[0].mem = bristolmalloc0(sizeof(float) * HISTSIZE);
	param->param[1].mem = bristolmalloc0(sizeof(float) * HISTSIZE);

	param->param[0].float_val = 0;
	param->param[1].float_val = 0;
	param->param[2].float_val = 0;
	param->param[3].float_val = 0;
	param->param[4].float_val = 1.0;
	param->param[5].float_val = 1.0;
}

static int operate(bristolOP *operator,
	bristolVoice *voice,
	bristolOPParams *param,
	void *lcl)
{
	bristolVREVERB *specs;
	bristolVREVERBlocal *local = (bristolVREVERBlocal *) lcl;
	register float *lsource, *rsource, *ldest, *rdest, *lhistory, *rhistory;
	register int count, i, histin, delay;
	register float wetdry, crossover, feedback, left, right;

	specs = (bristolVREVERB *) operator->specs;

	count = specs->spec.io[VREVERB_LOUT_IND].samplecount;
	lsource = specs->spec.io[VREVERB_LIN_IND].buf;
	rsource = specs->spec.io[VREVERB_RIN_IND].buf;
	ldest = specs->spec.io[VREVERB_LOUT_IND].buf;
	rdest = specs->spec.io[VREVERB_ROUT_IND].buf;

	/*
	 * operational parameters. Speed should be a function of depth, ie, if 
	 * depth changes, speed should be adjusted to maintain scan rate.
	 *
	 * Let speed be 10 seconds require to reach the given depth. That can be 
	 * calculated as 441000 samples, so we divide speed by depth * 44100;
	depth = param->param[DEPTH].float_val * 1024;
	speed = param->param[SPEED].float_val * 0.0002 * depth;
	 */
	/*
	 * Let scan be ten seconds to reach gain
	gain = param->param[GAIN].float_val * 1.5;
	scan = param->param[SCAN].float_val * 0.0005 * gain;
	 */
	delay = param->param[DELAY].float_val * HISTSIZE;
	feedback = param->param[FEEDBACK].float_val;
	crossover = param->param[CROSSOVER].float_val;
	wetdry = param->param[WETDRY].float_val;
	left = param->param[LEFT].float_val;
	right = param->param[RIGHT].float_val;

	lhistory = param->param[0].mem;
	rhistory = param->param[1].mem;

#ifdef DEBUG
	printf("reverb()\n");
#endif

	histin = local->histin;

	for (i = 0; i < count; i++) {
		/*
		 * Save our current signal into the history buffers. This should 
		 * include a component of feedback and crossover for echo/reverb effect.
		 */
		lhistory[histin] =
			lhistory[histin] * feedback
			+ rhistory[histin] * crossover
			+ *lsource++ * wetdry;
		rhistory[histin] =
			rhistory[histin] * feedback
			+ lhistory[histin] * crossover
			+ *rsource++ * wetdry;

		/*
		 * for testing just make the output be the input with delay.
		 */
		if (++histin >= delay)
			histin = 0;

		*ldest++ += lhistory[histin] * left;
		*rdest++ += rhistory[histin] * right;
	}

	local->histin = histin;
}

/*
 * Setup any variables in our OP structure, in our IO structures, and malloc
 * any memory we need.
 */
bristolOP *
reverbinit(bristolOP **operator, int index, int samplerate, int samplecount)
{
	bristolVREVERB *specs;

	*operator = bristolOPinit(operator, index, samplecount);

#ifdef BRISTOL_DBG
	printf("reverbinit(%x(%x), %i, %i, %i)\n",
		operator, *operator, index, samplerate, samplecount);
#endif

	/*
	 * Then the local parameters specific to this operator. These will be
	 * the same for each operator, but must be inited in the local code.
	 */
	(*operator)->operate = operate;
	(*operator)->destroy = destroy;
	(*operator)->reset = reset;
	(*operator)->param = param;
	(*operator)->flags |= BRISTOL_FX_STEREO;

	specs = (bristolVREVERB *) bristolmalloc0(sizeof(bristolVREVERB));
	(*operator)->specs = (bristolOPSpec *) specs;
	(*operator)->size = sizeof(bristolVREVERB);

	/*
	 * These are specific to this operator, and will need to be altered for
	 * each operator.
	 */
	specs->spec.opname = OPNAME;
	specs->spec.description = OPDESCRIPTION;
	specs->spec.pcount = PCOUNT;
	specs->spec.iocount = IOCOUNT;
	specs->spec.localsize = sizeof(bristolVREVERBlocal);

	/*
	 * Now fill in the specs for this operator.
	 */
	specs->spec.param[0].pname = "delay";
	specs->spec.param[0].description= "reverb delay line depth";
	specs->spec.param[0].type = BRISTOL_FLOAT;
	specs->spec.param[0].low = 0;
	specs->spec.param[0].high = 1;
	specs->spec.param[0].flags = BRISTOL_ROTARY|BRISTOL_SLIDER;

	specs->spec.param[1].pname = "feedback";
	specs->spec.param[1].description = "depth of mono feedback";
	specs->spec.param[1].type = BRISTOL_FLOAT;
	specs->spec.param[1].low = 0;
	specs->spec.param[1].high = 1;
	specs->spec.param[1].flags = BRISTOL_ROTARY|BRISTOL_SLIDER;

	specs->spec.param[2].pname = "crossfade";
	specs->spec.param[2].description = "depth of stereo feedback signal";
	specs->spec.param[2].type = BRISTOL_FLOAT;
	specs->spec.param[2].low = 0;
	specs->spec.param[2].high = 1;
	specs->spec.param[2].flags = BRISTOL_ROTARY|BRISTOL_SLIDER;

	specs->spec.param[3].pname = "depth";
	specs->spec.param[3].description = "wet/dry mix";
	specs->spec.param[3].type = BRISTOL_FLOAT;
	specs->spec.param[3].low = 0;
	specs->spec.param[3].high = 1;
	specs->spec.param[3].flags = BRISTOL_ROTARY|BRISTOL_SLIDER;

	specs->spec.param[4].pname = "left";
	specs->spec.param[4].description = "mix";
	specs->spec.param[4].type = BRISTOL_FLOAT;
	specs->spec.param[4].low = 0;
	specs->spec.param[4].high = 1;
	specs->spec.param[4].flags = BRISTOL_ROTARY|BRISTOL_SLIDER;

	specs->spec.param[5].pname = "right";
	specs->spec.param[5].description = "mix";
	specs->spec.param[5].type = BRISTOL_FLOAT;
	specs->spec.param[5].low = 0;
	specs->spec.param[5].high = 1;
	specs->spec.param[5].flags = BRISTOL_ROTARY|BRISTOL_SLIDER;

	/*
	 * Now fill in the IO specs.
	 */
	specs->spec.io[0].ioname = "left input buffer";
	specs->spec.io[0].description = "input signal";
	specs->spec.io[0].samplerate = samplerate;
	specs->spec.io[0].samplecount = samplecount;
	specs->spec.io[0].flags = BRISTOL_AC|BRISTOL_INPUT;

	specs->spec.io[1].ioname = "right input buffer";
	specs->spec.io[1].description = "input signal";
	specs->spec.io[1].samplerate = samplerate;
	specs->spec.io[1].samplecount = samplecount;
	specs->spec.io[1].flags = BRISTOL_AC|BRISTOL_INPUT;

	specs->spec.io[2].ioname = "left output";
	specs->spec.io[2].description = "output signal";
	specs->spec.io[2].samplerate = samplerate;
	specs->spec.io[2].samplecount = samplecount;
	specs->spec.io[2].flags = BRISTOL_AC|BRISTOL_OUTPUT;

	specs->spec.io[3].ioname = "right output";
	specs->spec.io[3].description = "output signal";
	specs->spec.io[3].samplerate = samplerate;
	specs->spec.io[3].samplecount = samplecount;
	specs->spec.io[3].flags = BRISTOL_AC|BRISTOL_OUTPUT;

	return(*operator);
}

